# -*- coding: utf-8 -*-
from PyQt5 import QtNetwork
from PyQt5.QtCore import Qt, QCoreApplication, QDateTime, QCoreApplication, QUrl, QVariant
from PyQt5.QtGui import QIcon, QPixmap, QColor, QFont
from PyQt5.QtWidgets import QDialog, QAction, QMessageBox, QFileDialog, QTableWidgetItem, QFrame, QComboBox, QCheckBox, QWidget, QHBoxLayout
from qgis.core import QgsProject, QgsWkbTypes, QgsMapLayer, QgsFeature, QgsGeometry, QgsPointXY, QgsVectorLayerSimpleLabeling, QgsPalLayerSettings, QgsTextBufferSettings, QgsTextFormat, QgsLabeling, QgsField
import string, math, os, sys

from .Check4SEC_dialog import Ui_Check4SEC_dialog


class Check4SEC(QDialog):

    def __init__(self, iface):
        self.z = ""

        super(Check4SEC, self).__init__()
        self.iface = iface

        if sys.platform.upper().startswith("WIN"):
            self.so = "W"
        else:
            self.so = "L"

        self.plugin_dir = self.manage_slash(os.path.dirname(__file__))
        self.project_dir = ""
        self.actions = ""
        self.menu = self.tr(u'&Check4SEC')
        self.toolbar = self.iface.addToolBar(u'Check4SEC')
        self.toolbar.setObjectName(u'Check4SEC')
        self.iface.newProjectCreated.connect(self.change_project)
        self.iface.projectRead.connect(self.change_project)

        self.create_parcel = False
        self.create_footprint = False
        


    def tr(self, message):
        return QCoreApplication.translate('Check4SEC', message)


    def manage_slash(self, bar):
        """ Manage slash '/', '\' """

        if bar.strip() == "" : return ""
        if self.so == "W":
            z = bar.replace('/', '\\')
            if str(z)[-1] == "\\" : z = z[0:-1]
        else :
            z = bar.replace('\\', '/')
            if str(z)[-1] == "/" : z = z[0:-1]

        return z


    def fill_zeros(self, v, n):
        """ Fill zeros """

        if len(str(v)) < n:
            return str("0000000000"[0:n-len(str(v))]) + str(v)
        else:
            return str(v)


    def initGui(self):
        self.init_vars()        
        # Config menu action
        icon = QIcon(self.manage_slash(self.plugin_dir + "/Check4SEC.png"))
        self.action = QAction(icon, "Check4SEC", self.iface.mainWindow())
        self.action.setObjectName("Check4SEC")
        self.action.setWhatsThis("Parcela Catastral GML")
        self.action.setStatusTip("Parcela Catastral GML")
        self.action.triggered.connect(self.run)

        # Config button
        self.iface.addToolBarIcon(self.action)
        self.iface.addPluginToMenu("&Check4SEC", self.action)
        
        # Config form
        self.ui = Ui_Check4SEC_dialog()
        self.ui.setupUi(self)
        self.ui.options_IVGA.setText('')
        self.ui.Logo.setPixmap(QPixmap(self.manage_slash(self.plugin_dir + "/Check4SEC.png")))
        self.ui.tabs.currentChanged.connect(lambda: self.validate_layer_features(True))
        self.ui.Validar_IVGA.clicked.connect(self.WFS_requests_parcel_if_connected)
        self.ui.Crear_IVGA.clicked.connect(self.change_create_parcel_flag)
        self.ui.Validar_ICUC.clicked.connect(self.WFS_requests_footprint_if_connected)
        self.ui.Crear_ICUC.clicked.connect(self.change_create_footprint_flag)
        self.ui.Carpeta.setIcon(QIcon(self.manage_slash(self.plugin_dir + "/folder.png")))
        self.ui.Carpeta.clicked.connect(self.select_folder)
        self.ui.Idioma.currentIndexChanged.connect(self.ui.retranslateUi)
        self.ui.Cerrar.clicked.connect(self.close) 
        
    def unload(self):
        # Remove icons and actions
        self.iface.removePluginMenu(self.tr(u'&Check4SEC'),self.action)
        self.iface.removeToolBarIcon(self.action)
        self.iface.newProjectCreated.disconnect(self.change_project)
        self.iface.projectRead.disconnect(self.change_project)
        self.ui.Validar_IVGA.clicked.disconnect(self.WFS_requests_parcel_if_connected)
        self.ui.Crear_IVGA.clicked.disconnect(self.change_create_parcel_flag)
        self.ui.Validar_ICUC.clicked.disconnect(self.WFS_requests_footprint_if_connected)
        self.ui.Crear_ICUC.clicked.disconnect(self.change_create_footprint_flag)
        self.ui.Idioma.currentIndexChanged.disconnect(self.ui.retranslateUi)
        self.ui.Carpeta.clicked.disconnect(self.select_folder)
        self.ui.Cerrar.clicked.disconnect(self.close)
        del self.toolbar

            
    def show_message(self, av, t):
        """ Show message to user """
        m = QMessageBox()
        m.setText(self.tr(t))
        if av == "P":
            m.setIcon(QMessageBox.Information)
            m.setWindowTitle("Pregunta")
            m.setStandardButtons(QMessageBox.Ok | QMessageBox.No)
            b1 = m.button(QMessageBox.Ok)
            b1.setText("Si")
            b2 = m.button(QMessageBox.No)
            b2.setText("No")
        else:
            if av == "W":
                m.setIcon(QMessageBox.Warning)
                z = "Atención"
            elif av == "C":
                m.setIcon(QMessageBox.Critical)
                z = "Error"
            else:
                m.setIcon(QMessageBox.Information)
            m.setWindowTitle("Aviso")
            m.setStandardButtons(QMessageBox.Ok)
            b = m.button(QMessageBox.Ok)
            b.setText("Seguir")

        return m.exec_()

    def console_out(self, line):
        self.ui.console.appendPlainText(line)


    def change_project(self):
        """ Update project folder """

        self.project_dir = self.manage_slash(QgsProject.instance().homePath())


    def select_folder(self):
        """ Select destination folder """

        msg = "Selección carpeta destino"
        z = QFileDialog.getExistingDirectory(None, msg, self.ui.desti.text(), QFileDialog.ShowDirsOnly)
        if z.strip() != "":
            self.ui.desti.setText(self.manage_slash(z))


    ##################### CHECKS BEFORE INIT #####################


    def askForConnection(self):
        if self.show_message("P", self.ui.msg_internet) == QMessageBox.No:
            self.show_message("W", self.ui.msg_noInternet)
            self.ui.check.setChecked(False)
            self.connected = False
            return False
        else:
            self.ui.check.setChecked(True)
            self.connected = True
            return True


    def validate_layer(self):
        """ Validate selected features of active layer """
        if not self.active_layer:
            self.show_message("C", self.ui.msg_noLayer)
            return False
        elif not self.active_layer.type() == QgsMapLayer.VectorLayer:
            self.show_message("C", self.ui.msg_noVector)
            return False
        else:
            self.iface.activeLayer().selectionChanged.connect(self.close)
            self.iface.layerTreeView().currentLayerChanged.connect(self.close)

        valid = ['EPSG:32628', 'EPSG:25829', 'EPSG:25830', 'EPSG:25831']
        self.crs = self.active_layer.crs().authid()
        if self.crs not in valid:
            self.show_message("C", self.ui.msg_noValidCRS)
            return False
        return True


    def validate_layer_features(self, show):
        self.elems = list(sorted(self.iface.activeLayer().selectedFeatures(), key = lambda feature: feature.geometry().area(), reverse=True))
        if self.ui.tabs.currentIndex() == 0:
            self.ui.Inspire4.setText("v4")
            if self.ui.tabs.isTabEnabled(0):
                if not self.elems or len(self.elems) == 0:
                    if show: self.show_message("C", self.ui.msg_noPolygon)
                    return False
                elif not any(elem.geometry().type() == QgsWkbTypes.PolygonGeometry for elem in self.elems):
                    if show: self.show_message("W", self.ui.msg_noPolygonSelected)
                    return False
                else:
                    self.ui.IVGA.setDisabled(False)
                    self.ui.tabs.setCurrentIndex(0)
                    fields = self.active_layer.fields().names()
                    self.change_combobox(self.ui.combobox_RC, fields)
                    if "REFCAT" in fields:
                        self.ui.combobox_RC.setCurrentText("REFCAT")
                    elif "refcat" in fields:
                        self.ui.combobox_RC.setCurrentText( "refcat")
                    elif "nationalCadastralReference" in fields:
                        self.ui.combobox_RC.setCurrentText("nationalCadastralReference")
                    elif "nationalCa" in fields:
                        self.ui.combobox_RC.setCurrentText("nationalCa")
                    elif "localId" in fields:
                        self.ui.combobox_RC.setCurrentText("localId")
                    return True
            else:
                return False
        else:
            names = [layer.name() for layer in QgsProject.instance().mapLayers().values() if layer.type() == QgsMapLayer.VectorLayer]
            self.change_combobox(self.ui.combobox_parcels, names)
            self.change_combobox(self.ui.combobox_footprint, names)
            self.ui.Inspire4.setText("v2")
    

    ##################### INIT GUI #####################
    def init_vars(self):
        self.connected = True
        self.active_layer = self.iface.activeLayer()
        self.elems = None
        self.isIVGA = False
        self.isICUC = False
        self.elems_wfs = []
        self.numberOfRequests = 0
        self.isBboxSplitted = False
        self.refs_to_check = []
        self.refs_not_found = []
        self.finishedRequests = 0
        self.equalParcels = []
        self.NPO = []
        self.errors = []

    def run(self):
        self.init_vars()
        self.ui.console.clear()
        self.ui.data.setDateTime(QDateTime.currentDateTime())
        # Validate selected features of active layer
        if self.validate_layer():
            names = [layer.name() for layer in QgsProject.instance().mapLayers().values() if layer.type() == QgsMapLayer.VectorLayer]
            self.change_combobox(self.ui.combobox_parcels, names)
            self.ui.combobox_parcels.currentTextChanged.connect(lambda: self.fill_ICUC_table("P"))
            self.change_combobox(self.ui.combobox_footprint, names)
            self.ui.combobox_footprint.currentTextChanged.connect(lambda: self.fill_ICUC_table("F"))
            self.fill_ICUC_table("P")
            self.fill_ICUC_table("F")

            """ We only activate IVGA tool if there's any feature selected in the active layer """
            self.ui.combobox_RC.currentTextChanged.connect(self.fill_IVGA_table)
            if not self.validate_layer_features(False):
                self.ui.IVGA.setDisabled(True)
                self.ui.tabs.setCurrentIndex(1)
                

            self.askForConnection()
            self.show()


    def change_combobox(self, combobox, array):
        combobox.clear()
        combobox.addItems(array)
        combobox.resize(combobox.sizeHint())
        

    def generate_labels(self, layer, elems):
        """ Create new attribute field in the layer for the geometry index """
        if layer.startEditing():
            if layer.dataProvider().fieldNameIndex("S") == -1:
                layer.dataProvider().addAttributes([QgsField("S", QVariant.String)])
                layer.updateFields()

            """ For each feature, set the index in the attribute field """
            col_name= layer.dataProvider().fieldNameIndex("S")
            for feature in layer.getFeatures():
                found = False
                for index, elem in enumerate(elems):
                    if feature.geometry().isGeosEqual(elem.geometry()):
                        layer.changeAttributeValue(feature.id(), col_name, str(index))
                        found = True
                if not found:
                    layer.changeAttributeValue(feature.id(), col_name, ' ')

            layer.commitChanges()
            
            """ Create the label to relate the parcel's geometry to an index in the table"""
            layer_settings = QgsPalLayerSettings()
            buffer_settings = QgsTextBufferSettings()
            buffer_settings.setEnabled(True)
            buffer_settings.setSize(1)
            buffer_settings.setColor(QColor("white"))

            text_format = QgsTextFormat()
            text_format.setFont(QFont("Arial", 6))
            text_format.setSize(6)

            text_format.setBuffer(buffer_settings)
            layer_settings.setFormat(text_format)
            layer_settings.fieldName = "S"
            layer_settings.placement = QgsPalLayerSettings.Curved
            layer_settings.placementFlags = QgsLabeling.OnLine
            layer_settings = QgsVectorLayerSimpleLabeling(layer_settings)

            layer.setLabelsEnabled(True)
            layer.setLabeling(layer_settings)
        
        else:
            #self.show_message("W", self.ui.msg_noEditable.format(layer.sourceName()))
            self.console_out("La capa {} no es editable, por lo que no se han podido añadir etiquetas.".format(layer.sourceName()))


    def fill_IVGA_table(self):
        if self.elems:
            nRef = ""
            if self.ui.combobox_RC.currentText() != None:
                nRef = self.ui.combobox_RC.currentText() 
            elif self.active_layer.fields().indexFromName("REFCAT") != -1:
                nRef = "REFCAT"
            elif self.active_layer.fields().indexFromName("refcat") != -1:
                nRef = "refcat"
            elif self.active_layer.fields().indexFromName("nationalCadastralReference") != -1:
                nRef = "nationalCadastralReference"
            elif self.active_layer.fields().indexFromName("nationalCa") != -1:
                nRef = "nationalCa"
            elif self.active_layer.fields().indexFromName("localId") != -1:
                nRef = "localId"

            self.ui.Table_IVGA.clear()
            fields = ["S", "RefCat", "Conservar", "LocalId", "Area", "Namespace"]
            self.ui.Table_IVGA.setColumnCount(len(fields))
            self.ui.Table_IVGA.setRowCount(len(self.elems))

            self.ui.Table_IVGA.setHorizontalHeaderLabels(fields)
            self.ui.Table_IVGA.horizontalHeader().setFrameStyle(QFrame.Box | QFrame.Plain)
            self.ui.Table_IVGA.horizontalHeader().setLineWidth(1)

            refs = []
            for index, elem in enumerate(self.elems):
                localId = self.generate_localId(index)
                area = int(elem.geometry().area())
                self.ui.Table_IVGA.setItem(index, 0, QTableWidgetItem(str(index)))
                self.ui.Table_IVGA.item(index, 0).setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
                
                ref = ""
                if nRef != "":
                    ref = elem[nRef]
                
                if ref != "":
                    if ref not in refs:
                        self.ui.Table_IVGA.setItem(index, 1, QTableWidgetItem(str(ref)))
                        refs.append(ref)
                    else:
                        self.ui.Table_IVGA.setItem(index, 1, QTableWidgetItem('NuevaRefCat_' + str(index)))
                else:
                    self.ui.Table_IVGA.setItem(index, 1, QTableWidgetItem(""))

                cell_widget = Widget()
                self.ui.Table_IVGA.setCellWidget(index, 2, cell_widget)

                self.ui.Table_IVGA.setItem(index, 3, QTableWidgetItem(str(localId)))
                self.ui.Table_IVGA.item(index, 3).setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
                c = QTableWidgetItem(str(area))
                c.setTextAlignment(Qt.AlignRight)
                self.ui.Table_IVGA.setItem(index, 4, c)
                self.ui.Table_IVGA.item(index, 4).setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
                cb1 = QComboBox()
                cb1.addItems(['LOCAL', 'SDGC'])
                self.ui.Table_IVGA.setCellWidget(index, 5, cb1)
                
                ref = self.ui.Table_IVGA.item(index,1).text()
                if nRef != "" and len(ref) == 14 and ref.isalnum():
                    self.ui.Table_IVGA.cellWidget(index, 2).setChecked(True)
                    self.ui.Table_IVGA.setItem(index, 3, QTableWidgetItem(str(ref)))
                    self.ui.Table_IVGA.cellWidget(index, 5).setCurrentIndex(1)
                else:    
                    self.ui.Table_IVGA.cellWidget(index, 2).setChecked(False)

            self.generate_labels(self.active_layer, self.elems)

            if self.ui.desti.text().strip() == "":
                if self.project_dir != "":
                    self.ui.desti.setText(self.project_dir)
                else:
                    self.ui.desti.setText(self.plugin_dir)
                

            self.ui.Table_IVGA.resizeRowsToContents()
            self.ui.Table_IVGA.resizeColumnsToContents()

    
    def fill_ICUC_table(self, type):
        combobox = self.ui.combobox_parcels if type == "P" else self.ui.combobox_footprint
        table = self.ui.Table_ICUC_parcels if type == "P" else self.ui.Table_ICUC_footprints
        table.clear()
        table.setRowCount(0)
        table.setColumnCount(0)
        layers = [lyr for lyr in QgsProject.instance().mapLayers().values() if lyr.name() == combobox.currentText()]
        if len(layers) > 0:
            layer = layers[0]
            elems = list(layer.getFeatures())
            self.generate_labels(QgsProject.instance().mapLayersByName(combobox.currentText())[0], elems)
            
            fields = layer.fields().names()
            if "S" in fields: fields.remove("S")
            fields.insert(0, "S")
            n=1
            if type == "F":
                fields.insert(1, "Seleccionar")
                n += 1
                self.change_combobox(self.ui.combobox_type, fields)
                self.change_combobox(self.ui.combobox_footprint_RC, fields)
                self.change_combobox(self.ui.combobox_floorsAbove, fields)
            
            table.setColumnCount(len(fields))
            table.setRowCount(len(elems))
            table.setHorizontalHeaderLabels(fields)
            table.horizontalHeader().setFrameStyle(QFrame.Box | QFrame.Plain)
            table.horizontalHeader().setLineWidth(1)
                
            for fil, elem in enumerate(elems):
                table.setItem(fil, 0, QTableWidgetItem(str(fil)))
                table.item(fil, 0).setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
                if type == "F":
                    cell_widget = Widget()
                    table.setCellWidget(fil, 1, cell_widget)
                for col in range(n, len(fields)):
                    table.setItem(fil, col, QTableWidgetItem(str(elem.attribute(col - n))))
                    table.item(fil, col).setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
            
            table.resizeRowsToContents()
            table.resizeColumnsToContents()
            if type == "P": 
                table.setStyleSheet("selection-background-color : orange; selection-color : white")
            


    ##################### COMMON UTILS #####################
    def request_bbox(self, xmin, ymin, xmax, ymax):
        self.manager_bbox = QtNetwork.QNetworkAccessManager()
        self.manager_bbox.finished.connect(self.handleResponse_bbox)
        
        bbox = '{},{},{},{}'.format(xmin, ymin, xmax, ymax)
        url = 'http://ovc.catastro.meh.es/INSPIRE/wfsCP.aspx?service=wfs&request=getfeature&Typenames=cp.cadastralparcel&SRSname=EPSG::{}&bbox={}'.format(str(self.crs.split(":")[1]), bbox)
        self.console_out('Petición BBOX: {}'.format(url))
        req = QtNetwork.QNetworkRequest(QUrl(url))
        self.manager_bbox.get(req)

    
    def split_bbox(self):
        self.console_out("El bbox es demasiado extenso, dividiendo petición...")
        self.isBboxSplitted = True
        bbox = self.active_layer.boundingBoxOfSelected()
        xmin, ymin, xmax, ymax = bbox.xMinimum(), bbox.yMinimum(), bbox.xMaximum(), bbox.yMaximum()
        dx, dy = 1000, 1000

        cols_int = math.ceil( (xmax - xmin)/dx )
        fils_int = math.ceil( (ymax - ymin)/dy )
        grid = []
        for fil in range(int(fils_int) + 1):
            grid_fil = []
            for col in range(int(cols_int) + 1):
                x = xmin + col * dx if (xmin + col * dx) < xmax else xmax
                y = ymin + fil * dy if (ymin + fil * dy) < ymax else ymax
                grid_fil.append([x, y])
            grid.insert(0, grid_fil)

        self.numberOfRequests += (len(grid) - 1) * (len(grid[0]) - 1)
        self.manager_bbox = QtNetwork.QNetworkAccessManager()
        self.manager_bbox.finished.connect(self.handleResponse_bbox)
        for i, fil in enumerate(grid):
            for j, col in enumerate(fil):
                if ( i < len(grid) - 1 and j < len(fil) - 1):
                    pmin, pmax = grid[i+1][j], grid[i][j+1]
                    xmin, ymin = pmin[0], pmin[1]
                    xmax, ymax = pmax[0], pmax[1]
                    bbox = '{},{},{},{}'.format(xmin, ymin, xmax, ymax)
                    url = 'http://ovc.catastro.meh.es/INSPIRE/wfsCP.aspx?service=wfs&request=getfeature&Typenames=cp.cadastralparcel&SRSname=EPSG::{}&bbox={}'.format(str(self.crs.split(":")[1]), bbox)
                    self.console_out('  Petición BBOX: {}'.format(url))
                    
                    req = QtNetwork.QNetworkRequest(QUrl(url))
                    self.manager_bbox.get(req)

    
    def handleResponse_bbox(self, reply):
        """ Parse WFS response to detect geometries and convert them to QGIS geometries """
        er = reply.error()
        if er == QtNetwork.QNetworkReply.NetworkError.NoError:
            bytes_string = reply.readAll()
            response = str(bytes_string, 'utf-8')
            #print(response)
            if "APLICACION EN MANTENIMIENTO" in response and "APLICACION EN MANTENIMIENTO" not in ";".join(self.errors):
                self.errors.append(self.ui.msg_WFSError.format("APLICACION EN MANTENIMIENTO"))
            elif 'Exception' in response:
                start = response.index('<ExceptionText><![CDATA[') + len('<ExceptionText><![CDATA[')
                end = response.index(']]></ExceptionText>')
                msg = response[start:end]
                if msg == "Area of extension out of limits":
                    if not self.isBboxSplitted:
                        self.split_bbox()
                    else:
                        if msg not in self.errors:
                            self.errors.append(self.ui.msg_WFSError.format(msg))
                
            else:
                cadastralParcels = [i for i in range(len(response)) if response.startswith('<cp:CadastralParcel gml:id="ES.SDGC.CP.',i)]
                rc = []
                for i in cadastralParcels:
                    rc.append(response[i + len('<cp:CadastralParcel gml:id="ES.SDGC.CP.'): i + len('<cp:CadastralParcel gml:id="ES.SDGC.CP.') + 14])

                cadastralParcelsEnd = [i for i in range(len(response)) if response.startswith('</cp:CadastralParcel>', i)]
                for index, i  in enumerate(cadastralParcels):
                    i += len('<cp:CadastralParcel gml:id="ES.SDGC.CP.XXXXXXXXXXXXXX">')
                    cp = response[i:cadastralParcelsEnd[index]]

                    polygonPatches = [j for j in range(len(cp)) if cp.startswith('<gml:PolygonPatch>', j)]
                    polygonPatchesEnd = [j for j in range(len(cp)) if cp.startswith('</gml:PolygonPatch>', j)]
                    for index1, j in enumerate(polygonPatches):

                        j += len('<gml:PolygonPatch>')
                        polygon = cp[j:polygonPatchesEnd[index1]]
                        exterior = polygon.find('<gml:exterior>') + len('<gml:exterior>')
                        exteriorEnd = polygon.find('</gml:exterior>')
                        exteriorRing = polygon[exterior:exteriorEnd]
                        posList = exteriorRing.find('<gml:posList') + len('<gml:posList')
                        n = exteriorRing[posList:].find('>')
                        posList = posList + n + len('>')
                        posListEnd = exteriorRing.find('</gml:posList>')
                        coordList_ext = exteriorRing[posList:posListEnd]             
                        
                        interiorRings = [k for k in range(len(polygon)) if polygon.startswith('<gml:interior>', k)]
                        interiorRingsEnd = [k for k in range(len(polygon)) if polygon.startswith('</gml:interior>', k)]
                        coordList_int = []
                        for index2, k in enumerate(interiorRings):
                            k += len('<gml:interior>')
                            interiorRing = polygon[k:interiorRingsEnd[index2]]
                            posList1 = interiorRing.find('<gml:posList') + len('<gml:posList')
                            n = interiorRing[posList1:].find('>')
                            posList1 = posList1 + n + len('>')
                            posList1End = interiorRing.find('</gml:posList>')
                            coordList_int.append(interiorRing[posList1:posList1End])

                        self.create_geometry(coordList_ext.split(), coordList_int, rc[index])
                
        self.check_finishedRequests()

    def check_finishedRequests(self):
        self.finishedRequests +=1
        self.console_out("Peticiones completadas: {}/{}".format(self.finishedRequests, self.numberOfRequests))
        if self.finishedRequests == self.numberOfRequests:
            if len(self.elems_wfs) > 0:
                if self.isIVGA:
                    self.validations_IVGA()
                    if self.create_parcel:
                        self.create_gml_parcel()
                    self.display_options()
                if self.isICUC:
                    self.validations_ICUC()
                    if self.create_footprint:
                        self.create_gml_footprint()
                    self.showErrors()
            else:
                self.showErrors()


    def create_geometry(self, coords_ext, rings_int, rc):
        """ Create polygons from the WFS response """
        if not any(rc in elem for elem in self.elems_wfs):
            polygon = []
            listOfPoints = []
            for index, coord in enumerate(coords_ext):
                if not index % 2:
                    listOfPoints.append((float(coord),float(coords_ext[index + 1])))
            polygon = [QgsPointXY( pair[0], pair[1] ) for pair in listOfPoints ]
            geometry = QgsGeometry.fromPolygonXY( [polygon] ) 

            if len(rings_int) > 0:
                for ring in rings_int:
                    listOfPoints = []
                    for index, coord in enumerate(ring.split()):
                        if not index % 2:
                            listOfPoints.append((float(coord),float(ring.split()[index + 1])))
                    part = [QgsPointXY( pair[0], pair[1] ) for pair in listOfPoints]
                    geometry.addRing(part)
            
            feature = QgsFeature()
            feature.setGeometry(geometry)
            
            self.elems_wfs.append([feature, rc])


    def get_rings(self, geom):
        """ Get list of rings of selected polygon """
        exterior_ring = []
        interior_rings = []
        wkb_type = geom.wkbType()
        if wkb_type == QgsWkbTypes.Polygon:
            polygon = geom.asPolygon()
            exterior_ring = polygon[0]

        elif wkb_type == QgsWkbTypes.MultiPolygon:
            list_polygons = geom.asMultiPolygon()
            polygon = list_polygons[0]
            exterior_ring = polygon[0]
            if len(polygon) > 1:
                for index, ring in enumerate(polygon):
                    if index > 0:
                        interior_rings.append(ring)

        return [exterior_ring, interior_rings]


    def get_points(self, ring):
        """ Get list of points of selected ring """
        punL = ""
        punN = len(ring)
        for point in ring:
            if punL != "":
                punL += " "
            punL += str("{:.2f}".format(round(point.x(),2))) + " " + str("{:.2f}".format(round(point.y(),2)))

        return [punN, punL]


    def translate_rings_to_gml(self, rings):
        exterior_ring = rings[0]
        [punN, punL] = self.get_points(exterior_ring)
        if len(rings) > 1:
            interior_rings = rings[1]
        txt='<gml:exterior>\n<gml:LinearRing>\n'
        txt+='<gml:posList srsDimension="2">'+punL+'</gml:posList>'
        txt+='\n</gml:LinearRing>\n</gml:exterior>'
        if len(interior_rings) > 0:
            for ring in interior_rings:
                [punN, punL] = self.get_points(ring)
                txt+= '\n<gml:interior>\n<gml:LinearRing>\n'
                txt+='<gml:posList srsDimension="2">'+punL+'</gml:posList>\n'
                txt+='</gml:LinearRing>\n</gml:interior>'
        return txt


    def generate_localId(self, index):
        """ Generate localId if the RefCat is not original """
        times = index/26
        frac, intNum = math.modf(times)
        frac = round(frac * 26, 0)
        try:
            localId = string.ascii_uppercase[int(intNum)] + string.ascii_uppercase[int(frac)]
            return localId
        except:
            return str(index)


    def check_overlaps(self, elems):
        valid = True
        self.console_out("Comprobando solapes entre las geometrías aportadas...")
        """ Validate that the selected parcels do not overlap each other """
        for index, elem in enumerate(elems):
            for index1, elem1 in enumerate(elems):
                if (index1 > index) and ( elem1.geometry().overlaps(elem.geometry()) or elem1.geometry().contains(elem.geometry()) or elem.geometry().contains(elem1.geometry()) ):
                    solape = elem.geometry().intersection(elem1.geometry()).area()
                    if solape > 0.9:
                        valid = False
                        self.console_out("  Solape entre las parcelas con índice {} y {}. Área de solape: {} m2".format(index, index1, solape))
        if valid:
            self.console_out("  OK")
        else:
            self.errors.append(self.ui.msg_overlappingFeatures)


    def compare_parcels_with_WFS(self):
        """ Validate that there are not original parcels without modifications in the selection, and count NPO """    
        self.console_out("Comparando parcelas aportadas con Catastro...")
        for index, elem in enumerate(self.elems):
            #GML Multipolygon to Polygon to match geometry types
            geom = None
            wkb_type = elem.geometry().wkbType()
            if wkb_type == QgsWkbTypes.MultiPolygon:
                for part in elem.geometry().asMultiPolygon():
                    geom = QgsGeometry.fromPolygonXY(part) if geom == None else geom.combine(QgsGeometry.fromPolygonXY(part) )
            elif wkb_type == QgsWkbTypes.Polygon:
                geom = elem.geometry() if geom == None else geom.combine(elem.geometry())

            for feature in self.elems_wfs:
                if geom.isGeosEqual(feature[0].geometry()):
                    if self.isIVGA: 
                        self.console_out("  COINCIDENCIA | Aportada: {}   Catastro: {}".format(self.ui.Table_IVGA.item(index, 1).text(), feature[1]))
                        self.equalParcels.append( 'Aportada: {}   Catastro: {}'.format(self.ui.Table_IVGA.item(index, 1).text(), feature[1]) )
                    elif self.isICUC:
                        self.equalParcels.append("YES")
                if ( (geom.isGeosEqual(feature[0].geometry()) or feature[0].geometry().contains(geom) or geom.contains(feature[0].geometry()) or geom.overlaps(feature[0].geometry())) and self.NPO.count(feature[1]) == 0 ):
                    self.NPO.append(feature[1])

        if self.isIVGA:
            if len(self.equalParcels) > 0:
                self.errors.append(self.ui.msg_equalGeometries.format('\n'.join(self.equalParcels)))
            self.console_out('  Parcelas Catastrales de origen: {}'.format(self.NPO))
        elif self.isICUC:
            if len(self.equalParcels) == 1:
                self.console_out("  OK")
            else:
                self.console_out("  NOK")
                self.errors.append(self.ui.msg_contourError_ICUC)


    def showErrors(self):
        self.console_out("-- PROCESO FINALIZADO --")
        self.isBboxSplitted = False
        if len(self.errors) > 0:
            self.show_message('C', "\n\n".join(self.errors))
            self.errors = []
            
        if not self.connected or (self.isIVGA and len(self.NPO) > 0) or self.isICUC:
            self.show_message('W', self.ui.msg_success)
        else:
            self.show_message('W', self.ui.msg_failed)

        self.init_vars()

    
    ##################### IVGA PROCESS #####################
    def change_create_parcel_flag(self):
        self.create_parcel = True
        self.WFS_requests_parcel_if_connected()


    def create_gml_parcel(self):
        """ Create GML file with Format INSPIRE """
        self.z = self.header_gml_parcel()
        self.write_gml_parcel()
        

    def write_gml_parcel(self):
        if (self.ui.check.isChecked() and not len(self.equalParcels) == self.NPP) or not self.ui.check.isChecked():
            # Check filename
            filename = str(self.ui.fileName.text())
            if filename == "":
                if self.show_message("W", self.ui.msg_noName) == QMessageBox.Ok:
                    return
            else:
                i = 0
                f = self.manage_slash(self.ui.desti.text() + "/" + filename + ".gml")
                if self.show_message("P", self.ui.msg_fileExists.format(filename)) == QMessageBox.No:
                    return
                else:
                    while os.path.exists(f):
                        i += 1
                        f = self.manage_slash(self.ui.desti.text() + "/" + filename + "_" + str(i) + ".gml")


            total = self.ui.Table_IVGA.rowCount()
            for fil in range(total):
                area = self.ui.Table_IVGA.item(fil, 4).text()
                namespace = self.ui.Table_IVGA.cellWidget(fil, 5).currentText()
                localId = ""
                if self.ui.Table_IVGA.cellWidget(fil, 2).isChecked():
                    self.ui.Table_IVGA.item(fil, 3).setText(self.ui.Table_IVGA.item(fil, 1).text())
                #Tags must not contain white spaces
                localId = self.ui.Table_IVGA.item(fil, 3).text().strip().replace(" ", "_")


                # Get list of points of selected polygon
                polygon = self.elems[fil].geometry()
                rings = self.get_rings(polygon)
                
                if len(rings) == 0:
                    return 
                epsg = self.crs.split(":")[1]
                
                self.z += self.body_gml_parcel(epsg, localId, area, namespace, rings)

            self.z += self.footer_gml_parcel()

            # Generate GML file
            fg = open(f, "w+")
            fg.write(self.z)
            fg.close()
            
            if i == 0:
                self.show_message("M", self.ui.msg_done.format(filename))
            else:
                self.show_message("M", self.ui.msg_done.format(filename + '_' + str(i)))
        else:
            self.show_message("C", self.ui.msg_noChanges)

        self.create_parcel = False


    def header_gml_parcel(self):
        """ GML header """
        z='<?xml version="1.0" encoding="utf-8"?>\n'
        z+='<!-- Archivo generado automaticamente por el plugin Check4SEC de QGIS. -->\n'
        z+='<!-- Parcela Catastral para entregar a la D.G. del Catastro. Formato INSPIRE v4. -->\n'
        z+='<FeatureCollection '
        z+='xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        z+='xmlns:gml="http://www.opengis.net/gml/3.2" '
        z+='xmlns:xlink="http://www.w3.org/1999/xlink" '
        z+='xmlns:cp="http://inspire.ec.europa.eu/schemas/cp/4.0" '
        z+='xmlns:gmd="http://www.isotc211.org/2005/gmd" '
        z+='xsi:schemaLocation="http://www.opengis.net/wfs/2.0 http://schemas.opengis.net/wfs/2.0/wfs.xsd http://inspire.ec.europa.eu/schemas/cp/4.0 http://inspire.ec.europa.eu/schemas/cp/4.0/CadastralParcels.xsd" '
        z+='xmlns="http://www.opengis.net/wfs/2.0" '
        z+='timeStamp="{}" '.format(self.ui.data.dateTime().toString(Qt.ISODate))
        z+='numberMatched="1" '
        z+='numberReturned="1">\n'

        return z


    def body_gml_parcel(self, epsg, localId, area, namespace, rings) :
        epsg = str(epsg)
        localId = str(localId)
        area = str(area)
        
        """ GML body """
        z='<member>\n' 
        z+='<cp:CadastralParcel gml:id="ES.'+namespace+'.CP.'+localId+'">\n'
        z+='<cp:areaValue uom="m2">'+area+'</cp:areaValue>\n'
        z+='<cp:beginLifespanVersion xsi:nil="true" nilReason="http://inspire.ec.europa.eu/codelist/VoidReasonValue/Unpopulated"></cp:beginLifespanVersion>\n'
        z+='<cp:endLifespanVersion xsi:nil="true" nilReason="http://inspire.ec.europa.eu/codelist/VoidReasonValue/Unpopulated"></cp:endLifespanVersion>\n'
        z+='<cp:geometry>\n'
        z+='<gml:MultiSurface gml:id="MultiSurface_ES.'+namespace+'.CP.'+localId+'" srsName="http://www.opengis.net/def/crs/EPSG/0/'+epsg+'">\n'
        z+='<gml:surfaceMember>\n'
        z+='<gml:Surface gml:id="Surface_ES.'+namespace+'.CP.'+localId+'.1" srsName="http://www.opengis.net/def/crs/EPSG/0/'+epsg+'">\n'
        z+='<gml:patches>\n<gml:PolygonPatch>\n'
        z+= self.translate_rings_to_gml(rings)
        z+='\n</gml:PolygonPatch>\n</gml:patches>\n'
        z+='</gml:Surface>\n</gml:surfaceMember>\n</gml:MultiSurface>\n</cp:geometry>\n'
        z+='<cp:inspireId>\n<Identifier xmlns="http://inspire.ec.europa.eu/schemas/base/3.3">\n'
        z+='<localId>'+localId+'</localId>\n<namespace>ES.'+namespace+'.CP</namespace>\n</Identifier>\n</cp:inspireId>\n'
        z+='<cp:label/>\n<cp:nationalCadastralReference/>'
        z+='\n</cp:CadastralParcel>\n'
        z+='</member>\n'

        return z


    def footer_gml_parcel(self) :
        """ GML footer """
        z = '</FeatureCollection>\n'

        return z

        
    def WFS_requests_parcel_if_connected(self):
        self.isIVGA = True
        self.elems = list(sorted(self.iface.activeLayer().selectedFeatures(), key = lambda feature: feature.geometry().area(), reverse=True))
        if self.ui.check.isChecked() or self.askForConnection(): 
            self.WFS_requests()
        else: 
            self.check_overlaps(self.elems)
            if self.create_parcel:
                self.create_gml_parcel()
            self.showErrors()
    

    def WFS_requests(self):
        self.ui.console.clear()
        self.console_out("-- PROCESO INICIADO --")
        self.numberOfRequests = 0

        """ WFS request: Get all geometries in the bbox """
        bbox = self.active_layer.boundingBoxOfSelected()
        xmin, ymin, xmax, ymax = bbox.xMinimum(), bbox.yMinimum(), bbox.xMaximum(), bbox.yMaximum()

        self.numberOfRequests += 1
        self.request_bbox(xmin, ymin, xmax, ymax)
        

        """ Review of all the given geometries, to look for the ones to check in the WFS"""
        self.NPP = self.ui.Table_IVGA.rowCount()
        for fil in range(self.NPP):
            ref = self.ui.Table_IVGA.item(fil, 1).text().upper()
            if len(ref) == 14 and ref.isalnum():
                self.refs_to_check.append([ref,fil])
            else:
                self.ui.Table_IVGA.cellWidget(fil, 5).setCurrentIndex(0)
            
            if self.ui.Table_IVGA.cellWidget(fil, 2).isChecked():
                self.ui.Table_IVGA.item(fil, 3).setText(self.ui.Table_IVGA.item(fil, 1).text())
            else:
                self.ui.Table_IVGA.item(fil, 3).setText(self.generate_localId(fil))

        """ WFS request: Cheack each RefCat for a match """
        self.manager_RC = QtNetwork.QNetworkAccessManager()
        self.manager_RC.finished.connect(self.handleResponse_RC)
        if len(self.refs_to_check) > 0:
            self.numberOfRequests += len(self.refs_to_check)
            for item in self.refs_to_check:
                url = 'http://ovc.catastro.meh.es/INSPIRE/wfsCP.aspx?service=wfs&version=2&request=getfeature&STOREDQUERIE_ID=GetParcel&refcat={}&srsname=EPSG::{}'.format(item[0],str(self.crs.split(":")[1]))
                self.console_out('Petición RC: {}'.format(url))
                req = QtNetwork.QNetworkRequest(QUrl(url))
                self.manager_RC.get(req)
        


    def handleResponse_RC(self, reply):
        """ For all the RefCat that want to be preserved, there is a validation against the WFS to verify that the Refcat exists """
        er = reply.error()
        if er == QtNetwork.QNetworkReply.NetworkError.NoError:
            bytes_string = reply.readAll()
            response = str(bytes_string, 'utf-8')
            #print(response)
            if "APLICACION EN MANTENIMIENTO" in response and "APLICACION EN MANTENIMIENTO" not in ";".join(self.errors):
                self.errors.append(self.ui.msg_WFSError.format("APLICACION EN MANTENIMIENTO"))
            elif 'Exception' in response:
                start = response.index('<ExceptionText><![CDATA[No se ha encontrado la parcela ') + len('<ExceptionText><![CDATA[No se ha encontrado la parcela ')
                end = start + 14
                ref = response[start:end]
                for i in range(len(self.refs_to_check)):
                    if self.refs_to_check[i][0] == ref:    
                        fil = self.refs_to_check[i][1]
                        if self.ui.Table_IVGA.cellWidget(fil, 2).isChecked():
                            self.refs_not_found.append(ref)
                        self.ui.Table_IVGA.cellWidget(fil, 5).setCurrentIndex(0)

            elif 'cp:CadastralParcel' in response:
                start = response.index('<localId>') + len('<localId>')
                end = start + 14
                ref = response[start:end]
                for i in range(len(self.refs_to_check)):
                    if self.refs_to_check[i][0] == ref:    
                        fil = self.refs_to_check[i][1]
                        if self.ui.Table_IVGA.cellWidget(fil, 2).isChecked():
                            self.ui.Table_IVGA.cellWidget(fil, 5).setCurrentIndex(1)
                        else:
                            self.ui.Table_IVGA.cellWidget(fil, 5).setCurrentIndex(0)
            
                        if fil == self.refs_to_check[-1][1]:
                            if len(self.refs_not_found) > 0 :
                                msg = ''
                                for ref in self.refs_not_found:
                                    msg += '\n' + ref
                                self.errors.append(self.ui.msg_notFound.format(msg))
            
            self.check_finishedRequests()
            

    def validations_IVGA(self):
        """ Validate that the selected parcels do not overlap each other """
        self.check_overlaps(self.elems)

        """ Validate that there are not original parcels without modifications in the selection, and count NPO """
        self.compare_parcels_with_WFS()    
                
        """ Validate that the contours for the original parcels and the modified parcels are the same """
        if len(self.NPO) > 0:
            self.compare_contours()
        else:
            self.console_out("  No se han encontrado parcelas de origen en Catastro")
            self.errors.append(self.ui.msg_noOriginParcelFound)


    def compare_contours(self):
        self.console_out("Comparando perímetros exteriores...")
        
        geom = None
        for feat in self.elems:
            wkb_type = feat.geometry().wkbType()
            if wkb_type == QgsWkbTypes.MultiPolygon:
                for part in feat.geometry().asMultiPolygon():
                    geom = QgsGeometry.fromPolygonXY(part) if geom == None else geom.combine(QgsGeometry.fromPolygonXY(part))
            elif wkb_type == QgsWkbTypes.Polygon:
                geom = feat.geometry() if geom == None else geom.combine(feat.geometry())
        
        geom_wfs = None
        for feat in self.elems_wfs:
            if self.NPO.count(feat[1]) > 0:
                geom_wfs = feat[0].geometry() if geom_wfs == None else geom_wfs.combine(feat[0].geometry())
        
        diferencia = abs(geom.area() - geom_wfs.area())
        if geom.isGeosEqual(geom_wfs) or diferencia < 0.9:
            self.console_out("  OK")
        else:
            self.console_out('  Diferencia de área: {}'.format(diferencia))
            if geom.area() == geom_wfs.area():
                diff_points = []  
                for i in range(len(geom.asPolygon()[0])):
                    if geom.asPolygon()[0][i] not in geom_wfs.asPolygon()[0]:
                        diff_points.append(geom.asPolygon()[0][i])
                if len(diff_points) > 0:
                    points = [point for point in geom.asPolygon()[0] if point not in diff_points]
                    geom_bis = QgsGeometry.fromPolygonXY( [points] ) 
                if not geom_bis.area() == geom_wfs.area() or not geom_bis.isGeosEqual(geom_wfs):
                    self.errors.append(self.ui.msg_contourError)
            else:
                self.console_out("  NOK")
                self.errors.append(self.ui.msg_contourError)
        


    def display_options(self):
        """ Show the proccedures available for the selected cadastral parcels """
        self.console_out("NPO: {}; NPP: {}".format(len(self.NPO), self.NPP))
        total = self.ui.Table_IVGA.rowCount()
        namespaces = []
        for fil in range(total):
            namespace = self.ui.Table_IVGA.cellWidget(fil, 5).currentIndex()
            namespaces.append(namespace)
        LOCAL = namespaces.count(0)
        SDGC = namespaces.count(1)
        if len(self.NPO) == 1 and self.NPP > 1:
            if SDGC == 0:
                output = 'División' if self.ui.Idioma.currentIndex() == 0 else 'Division'
            elif SDGC == 1:
                output = 'Segregación, Subsanación' if self.ui.Idioma.currentIndex() == 0 else 'Segregation, Rectification'
            elif SDGC > 1:
                output = 'NO PERMITIDO'    
        elif len(self.NPO) > 1 and self.NPP == 1:
            if SDGC == 0 and LOCAL == 1:
                output = 'Agrupación' if self.ui.Idioma.currentIndex() == 0 else 'Aggregation'
            if SDGC == 1:
                output = 'Agregación, Subsanación' if self.ui.Idioma.currentIndex() == 0 else 'Aggregation, Rectification'
        elif len(self.NPO) == self.NPP:
            if LOCAL == 0:
                output = 'Subsanación' if self.ui.Idioma.currentIndex() == 0 else 'Rectification'
            elif LOCAL > 0:
                output = 'Subsanación, Reparcelación' if self.ui.Idioma.currentIndex() == 0 else 'Rectification, Reparcelization'
        elif len(self.NPO) > 0 and len(self.NPO) != self.NPP:
            output = 'Subsanación, Reparcelación' if self.ui.Idioma.currentIndex() == 0 else 'Rectification, Reparcelization'
        else:
            output = ''
        self.console_out("Posibles alteraciones: {}".format(output))
        self.ui.options_IVGA.setText(output)  

        self.showErrors()


    ##################### ICUC PROCESS #####################
    def change_create_footprint_flag(self):
        self.create_footprint = True
        self.WFS_requests_footprint_if_connected()


    def create_gml_footprint(self):
        """ Create GML file with Format INSPIRE """
        self.z = self.header_gml_footprint()
        self.write_gml_footprint()
        self.create_footprint = False
    
    
    def write_gml_footprint(self):
        # Check filename
        filename = str(self.ui.fileName.text())
        if filename == "":
            if self.show_message("W", self.ui.msg_noName) == QMessageBox.Ok:
                return
        else:
            i = 0
            f = self.manage_slash(self.ui.desti.text() + "/" + filename + ".gml")
            if self.show_message("P", self.ui.msg_fileExists.format(filename)) == QMessageBox.No:
                return
            else:
                while os.path.exists(f):
                    i += 1
                    f = self.manage_slash(self.ui.desti.text() + "/" + filename + "_" + str(i) + ".gml")


        layer = [lyr for lyr in QgsProject.instance().mapLayers().values() if lyr.name() == self.ui.combobox_footprint.currentText()][0]
        rc = []
        numberPool = 0
        for row in range(self.ui.Table_ICUC_footprints.rowCount()):
            if self.ui.Table_ICUC_footprints.cellWidget(row, 1).isChecked():

                epsg = self.crs.split(":")[1]
                type_col = self.ui.combobox_type.currentIndex()
                type = self.ui.Table_ICUC_footprints.item(row, type_col).text()
                floors_col = self.ui.combobox_floorsAbove.currentIndex()
                floors = self.ui.Table_ICUC_footprints.item(row, floors_col).text()

                refcat_col = self.ui.combobox_footprint_RC.currentIndex()
                refcat = self.ui.Table_ICUC_footprints.item(row, refcat_col).text()
                if refcat not in rc:
                    localId = refcat
                else:
                    localId = self.generate_localId(row)

                namespace = "LOCAL"
                
                if type != "BU": numberPool+=1
                n = None if type == "BU" else numberPool

                # Get list of points of selected polygon
                polygon = list(layer.getFeatures())[row].geometry()
                area = polygon.area()
                rings = self.get_rings(polygon)
            
                if len(rings) == 0:
                    return 
                
                self.z += self.body_gml_footprint(epsg, localId, area, namespace, rings, type, floors, n)

        self.z += self.footer_gml_footprint()

        # Generate GML file
        fg = open(f, "w+")
        fg.write(self.z)
        fg.close()
        if i == 0:
            self.show_message("M", self.ui.msg_done.format(filename))
        else:
            self.show_message("M", self.ui.msg_done.format(filename + '_' + str(i)))

        self.create_footprint = False


    def header_gml_footprint(self):
        """ GML header """
        z='<?xml version="1.0" encoding="ISO-8859-1"?>\n'
        z+='<!-- Archivo generado automaticamente por el plugin Check4SEC de QGIS. -->\n'
        z+='<!-- Building OtherConstruction para entregar a la D.G. del Catastro. Formato INSPIRE v2. -->\n'
        z+='<gml:FeatureCollection gml:id="ES.LOCAL.BU" '
        z+= 'xmlns:ad="urn:x-inspire:specification:gmlas:Addresses:3.0" '
        z+='xmlns:base="urn:x-inspire:specification:gmlas:BaseTypes:3.2" '
        z+='xmlns:bubase="http://inspire.jrc.ec.europa.eu/schemas/bu-base/3.0" '
        z+='xmlns:bu-core2d="http://inspire.jrc.ec.europa.eu/schemas/bu-core2d/2.0" '
        z+='xmlns:bu-ext2d="http://inspire.jrc.ec.europa.eu/schemas/bu-ext2d/2.0" '
        z+='xmlns:cp="urn:xinspire:specification:gmlas:CadastralParcels:3.0" '
        z+='xmlns:el-bas="http://inspire.jrc.ec.europa.eu/schemas/elbas/2.0" '
        z+='xmlns:el-cov="http://inspire.jrc.ec.europa.eu/schemas/el-cov/2.0" '
        z+='xmlns:eltin="http://inspire.jrc.ec.europa.eu/schemas/el-tin/2.0" '
        z+='xmlns:elvec="http://inspire.jrc.ec.europa.eu/schemas/el-vec/2.0" '
        z+='xmlns:gco="http://www.isotc211.org/2005/gco" '
        z+='xmlns:gmd="http://www.isotc211.org/2005/gmd" '
        z+='xmlns:gml="http://www.opengis.net/gml/3.2" '
        z+='xmlns:gmlcov="http://www.opengis.net/gmlcov/1.0" '
        z+='xmlns:gn="urn:xinspire:specification:gmlas:GeographicalNames:3.0" '
        z+='xmlns:gsr="http://www.isotc211.org/2005/gsr" '
        z+='xmlns:gss="http://www.isotc211.org/2005/gss" '
        z+='xmlns:gts="http://www.isotc211.org/2005/gts" '
        z+='xmlns:swe="http://www.opengis.net/swe/2.0" '
        z+='xmlns:xlink="http://www.w3.org/1999/xlink" '
        z+='xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        z+='xsi:schemaLocation="http://inspire.jrc.ec.europa.eu/schemas/bu-ext2d/2.0 http://inspire.ec.europa.eu/draftschemas/bu-ext2d/2.0/BuildingExtended2D.xsd">\n'
        
        return z


    def body_gml_footprint(self, epsg, localId, area, namespace, rings, type, floors, n):
        epsg = str(epsg)
        localId = str(localId)
        area = str(area)
        
        """ GML body """
        construction = 'Building' if type == "BU" else 'OtherConstruction'

        z='<gml:featureMember>\n'
        z+='<bu-ext2d:'+construction+' gml:id="ES.SDGC.BU.'+localId+'">\n'
        z+='<bu-core2d:beginLifespanVersion>'+format(self.ui.data.dateTime().toString(Qt.ISODate))+'</bu-core2d:beginLifespanVersion>\n'
        z+= '<bu-core2d:conditionOfConstruction>functional</bu-core2d:conditionOfConstruction>\n' if type == "BU" else '<bu-core2d:conditionOfConstruction xsi:nil="true" nilReason="other:unpopulated"></bu-core2d:conditionOfConstruction>'
        z+='<bu-core2d:inspireId>\n'
        z+='<base:Identifier>\n'
        z+='<base:localId>'+localId+'</base:localId>\n'
        z+='<base:namespace>ES.'+namespace+'.BU</base:namespace>\n'
        z+='</base:Identifier>\n'
        z+='</bu-core2d:inspireId>\n'
        if type == "BU": 
            z+='<bu-ext2d:geometry>\n'
            z+='<bu-core2d:BuildingGeometry>\n'
            z+='<bu-core2d:geometry>\n'
            z+='<gml:Surface gml:id="Surface_ES.'+namespace+'.BU.'+localId+'_PI'+str(n)+'" srsName="urn:ogc:def:crs:EPSG::'+epsg+'">\n'
            z+='<gml:patches>\n<gml:PolygonPatch>\n'
            z+= self.translate_rings_to_gml(rings)
            z+='\n</gml:PolygonPatch>\n</gml:patches>\n'
            z+='</gml:Surface>\n'
            z+='</bu-core2d:geometry>\n'
            z+='<bu-core2d:horizontalGeometryEstimatedAccuracy uom="m">0.1</bu-core2d:horizontalGeometryEstimatedAccuracy>\n'
            z+='<bu-core2d:horizontalGeometryReference>footPrint</bu-core2d:horizontalGeometryReference>\n'
            z+='<bu-core2d:referenceGeometry>true</bu-core2d:referenceGeometry>\n'
            z+='</bu-core2d:BuildingGeometry>\n'
            z+='</bu-ext2d:geometry>\n'
            try:
                int(floors)
                z+='<bu-ext2d:numberOfFloorsAboveGround>'+floors+'</bu-ext2d:numberOfFloorsAboveGround>\n'
            except:
                z+='<bu-ext2d:numberOfFloorsAboveGround xsi:nil="true" nilReason="http://inspire.ec.europa.eu/codelist/VoidReasonValue/Unpopulated"></bu-ext2d:numberOfFloorsAboveGround>\n'
        else:
            z+='<bu-ext2d:constructionNature>openAirPool</bu-ext2d:constructionNature>\n'
            z+='<bu-ext2d:geometry>\n'
            z+='<gml:Polygon gml:id="Polygon_ES.'+namespace+'.BU.'+localId+'_PI.'+str(n)+'" srsName="urn:ogc:def:crs:EPSG::'+epsg+'">\n'
            z+= self.translate_rings_to_gml(rings)
            z+='\n'
            z+='</gml:Polygon>\n'
            z+='</bu-ext2d:geometry>\n'
        z+='</bu-ext2d:'+construction+'>\n'
        z+='</gml:featureMember>\n'

        return z

    
    def footer_gml_footprint(self):
        """ GML footer """
        z='</gml:FeatureCollection>'

        return z


    def WFS_requests_footprint_if_connected(self):
        self.isICUC = True
        if self.ui.check.isChecked() or self.askForConnection(): 
            self.WFS_requests_footprint()
        else:
            self.validations_ICUC()
            if self.create_footprint:
                self.create_gml_footprint()
            self.showErrors()
    

    def WFS_requests_footprint(self):
        self.ui.console.clear()
        self.numberOfRequests = 0
        rows = self.ui.Table_ICUC_parcels.selectionModel().selectedRows()
        
        if self.ui.selected_layer.isChecked() and len(rows) == 1:
            self.console_out("-- PROCESO INICIADO --")
            self.elems = [self.get_parcel(rows[0].row())]
            bbox = self.elems[0].geometry().boundingBox()
            xmin, ymin, xmax, ymax = bbox.xMinimum(), bbox.yMinimum(), bbox.xMaximum(), bbox.yMaximum()
            """ WFS request: Get all geometries in the bbox """
            self.numberOfRequests += 1
            self.request_bbox(xmin, ymin, xmax, ymax)
        elif self.ui.selected_rc.isChecked():
            rc = self.ui.parcel_rc.text()
            if len(rc) == 14 and rc.isalnum():
                self.console_out("-- PROCESO INICIADO --")
                self.numberOfRequests += 1
                self.manager_RC = QtNetwork.QNetworkAccessManager()
                self.manager_RC.finished.connect(self.handleResponse_bbox)
                url = 'http://ovc.catastro.meh.es/INSPIRE/wfsCP.aspx?service=wfs&version=2&request=getfeature&STOREDQUERIE_ID=GetParcel&refcat={}&srsname=EPSG::{}'.format(rc,str(self.crs.split(":")[1]))
                self.console_out('Petición RC: {}'.format(url))
                req = QtNetwork.QNetworkRequest(QUrl(url))
                self.manager_RC.get(req)
            else:
                self.show_message("C", self.ui.msg_noValidRC)


    def validations_ICUC(self):
        """ Validate construction type """
        self.check_type_field()

        """ Validate that the selected footprints do not overlap each other """
        self.check_overlaps(self.get_footprints()[0])
        
        """ Validate that the selected parcel contains the footprints """
        rows = self.ui.Table_ICUC_parcels.selectionModel().selectedRows()
        if self.ui.selected_layer.isChecked() and len(rows) == 1:
            self.console_out("Parcela seleccionada: S{}".format(rows[0].row()))
            self.parcel_contains_footprints(self.get_parcel(rows[0].row()), self.get_footprints())
            """ Validate that the parcel is equal to Cadastre, only if there's internet connection """
            if self.connected:
                self.compare_parcels_with_WFS()
        elif self.ui.selected_rc.isChecked() and self.connected and len(self.elems_wfs) ==1:
            self.console_out("Parcela seleccionada: {}".format(self.elems_wfs[0][1]))
            self.parcel_contains_footprints(self.elems_wfs[0][0], self.get_footprints())

    
    def check_type_field(self):
        self.console_out("Comprobando campo de tipo de construcción...")
        msg = None
        for row in range(self.ui.Table_ICUC_footprints.rowCount()):
            if self.ui.Table_ICUC_footprints.cellWidget(row, 1).isChecked():
                type_col = self.ui.combobox_type.currentIndex()
                type = self.ui.Table_ICUC_footprints.item(row, type_col).text()
                if type != "BU" and type != "PI":
                    msg += "\n S:{}".format(row)
                    self.console_out("La huella S:{} no tiene un tipo de construcción válido.".format(row))
        if msg:
            self.show_message('C', self.ui.msg_NoValidTypeConstruction)
        else:
            self.console_out("  OK")


    
    def parcel_contains_footprints(self, parcel, footprints):
        self.console_out("Comprobando si las huellas están contenidas en la parcela...")
        flag = True
        msg = ""
        for index, footprint in enumerate(footprints[0]):
            row = footprints[1][index]
            if parcel.geometry().contains(footprint.geometry()):
                self.change_row_color(row, QColor("green"))
            else:
                self.change_row_color(row, QColor("red"))
                self.console_out("La huella S:{} no se encuentra dentro de la parcela.".format(index))
                msg += "\n S:{}".format(row)
                flag = False
                
        if flag:
            self.console_out("  OK")
        else:
            self.errors.append(self.ui.msg_footprint.format(msg))

    def change_row_color(self, row, color):
        self.ui.Table_ICUC_footprints.item(row, 0).setBackground(color)


    def get_parcel(self, index):
        layer = [lyr for lyr in QgsProject.instance().mapLayers().values() if lyr.name() == self.ui.combobox_parcels.currentText()][0]
        parcel = list(layer.getFeatures())[index]
        return parcel


    def get_footprints(self):
        footprints = []
        rows = []
        layer = [lyr for lyr in QgsProject.instance().mapLayers().values() if lyr.name() == self.ui.combobox_footprint.currentText()][0]
        for row in range(self.ui.Table_ICUC_footprints.rowCount()):
            self.change_row_color(row, QColor("white"))
            if self.ui.Table_ICUC_footprints.cellWidget(row, 1).isChecked():
                footprints.append(list(layer.getFeatures())[row])
                rows.append(row)

        return [footprints, rows]    



##################### CUSTOM CHECKBOX CENTERED IN TABLE COLUMN #####################
class Widget(QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.chk_box = QCheckBox()
        self.chk_box.setCheckState(Qt.Unchecked)
        hlayout = QHBoxLayout(self)
        hlayout.addWidget(self.chk_box)
        hlayout.setAlignment(Qt.AlignCenter)
        hlayout.setContentsMargins(0, 0, 0, 0)

    def isChecked(self):
        return self.chk_box.isChecked()

    def setChecked(self, checked):
        self.chk_box.setChecked(checked)