from PyQt5 import QtNetwork
from PyQt5.QtCore import Qt, QCoreApplication, QDateTime, QCoreApplication, QUrl, QVariant
from PyQt5.QtGui import QIcon, QPixmap, QColor, QFont
from PyQt5.QtWidgets import QDialog, QAction, QMessageBox, QFileDialog, QTableWidgetItem, QFrame, QComboBox, QCheckBox, QWidget, QHBoxLayout, QRadioButton
from qgis.core import QgsProject, QgsWkbTypes, QgsMapLayer, QgsFeature, QgsGeometry, QgsPointXY, QgsVectorLayerSimpleLabeling, QgsPalLayerSettings, QgsTextBufferSettings, QgsTextFormat, QgsLabeling, QgsField
import string, math, os, sys

from .Check4SEC_dialog import Ui_Check4SEC_dialog


class IVGA(QDialog, Ui_Check4SEC_dialog):

    def __init__(self, parent):
        
        super(IVGA, self).__init__(parent == None)


    def fill_IVGA_table(self):
        # Validate layer fields
        nRef = self.look_for_refcat(self.layer)        

        # Load form list
        self.parent.ui.data.setDateTime(QDateTime.currentDateTime())
        self.parent.ui.Selec.clear()
        fields = ["S", "RefCat", "Conservar", "LocalId", "Area", "Namespace"]
        self.parent.ui.Selec.setColumnCount(len(fields))
        self.parent.ui.Selec.setRowCount(len(self.parent.elems))

        self.parent.ui.Selec.setHorizontalHeaderLabels(fields)
        self.parent.ui.Selec.horizontalHeader().setFrameStyle(QFrame.Box | QFrame.Plain)
        self.parent.ui.Selec.horizontalHeader().setLineWidth(1)

        self.geo = []
        refs = []
        for index, elem in enumerate(self.parent.elems):
            self.geo.append(elem.geometry())

            localId = self.parent.generate_localId(index)
            area = int(elem.geometry().area())
            ref = ""
            if nRef != "":
                ref = elem[nRef]

            self.parent.ui.Selec.setItem(index, 0, QTableWidgetItem(str(index)))
            self.parent.ui.Selec.item(index, 0).setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
            
            if ref != "":
                if ref not in refs:
                    self.parent.ui.Selec.setItem(index, 1, QTableWidgetItem(str(ref)))
                    refs.append(ref)
                else:
                    self.parent.ui.Selec.setItem(index, 1, QTableWidgetItem('NuevaRefCat_' + str(index)))
            else:
                self.parent.ui.Selec.setItem(index, 1, QTableWidgetItem(""))

            cell_widget = self.parent.Widget()
            self.parent.ui.Selec.setCellWidget(index, 2, cell_widget)

            self.parent.ui.Selec.setItem(index, 3, QTableWidgetItem(str(localId)))
            self.parent.ui.Selec.item(index, 3).setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
            c = QTableWidgetItem(str(area))
            c.setTextAlignment(Qt.AlignRight)
            self.parent.ui.Selec.setItem(index, 4, c)
            self.parent.ui.Selec.item(index, 4).setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
            cb1 = QComboBox()
            cb1.addItems(['LOCAL', 'SDGC'])
            self.parent.ui.Selec.setCellWidget(index, 5, cb1)
            ref = self.parent.ui.Selec.item(index,1).text()
            if nRef != "" and len(ref) == 14 and " " not in ref and "_" not in ref:
                self.parent.ui.Selec.cellWidget(index, 2).setChecked(True)
                self.parent.ui.Selec.setItem(index, 3, QTableWidgetItem(str(ref)))
                self.parent.ui.Selec.cellWidget(index, 5).setCurrentIndex(1)
            else:    
                self.parent.ui.Selec.cellWidget(index, 2).setChecked(False)

        self.parent.ui.Selec.resizeRowsToContents()
        self.parent.ui.Selec.resizeColumnsToContents()

        if self.parent.ui.desti.text().strip() == "":
            if self.parent.project_dir != "":
                self.parent.ui.desti.setText(self.project_dir)
            else:
                self.parent.ui.desti.setText(self.plugin_dir)

    
    